! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      
      implicit none
      real(dp) :: T0, T1, tt, dimtt, dimTime !Time variables
      real(dp) :: a0, b0, c0, prefac 
      real(dp) :: Heat, Time, eta, sma, ecc, Grav, Mbh, M, R, Edot, Edota, Edotb !Stellar and orbital eovlution variables
      real(dp) :: estart,dedt;
      real(dp) :: x1   

      integer :: time0, time1, clock_rate
      double precision, parameter :: expected_runtime = 6.8 ! minutes

      
      contains
      
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  
      end subroutine extras_controls
      
!Additional subroutines for calculating eccentricity evolution of the orbit due to gravitational radiation. RK4 is accurate for large eccentricities when the timestep is capped at 10-100 years. When the orbit is nearly circular, other elements of the tidal heating calculation are larger sources of uncertainty than error from orbital integration.

      subroutine rk4(e, t, h, eout) !equations from Peters 1964
      	implicit none
	real(dp), intent(in) :: e, t, h
	real(dp), intent(out) :: eout
	real(dp) :: dedt, tmid, de1, de2, emid
	dedt = prefac*e**(-29./19.)*(1.-e**2)**(1.5)/(1 + (121./304.)*e**2)**(1181./2299.)
	tmid=t+h*0.5_dp
	emid=e+h*0.5_dp*dedt
	de1 = prefac*emid**(-29./19.)*(1.-emid**2)**(1.5)/(1 + (121./304.)*emid**2)**(1181./2299.)
	emid = e+h*0.5_dp*de1
	de2 = prefac*emid**(-29./19.)*(1.-emid**2)**(1.5)/(1 + (121./304.)*emid**2)**(1181./2299.)
	emid = e+ h*de2
	de2 = de1+de2
	de1 = prefac*emid**(-29./19.)*(1.-emid**2)**(1.5)/(1 + (121./304.)*emid**2)**(1181./2299.)
	eout = e+h*(dedt+de1+2.0_dp*de2)/6.0_dp 
      end subroutine rk4

      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)


	 dimTime = 0. 
	 Heat = 0. !Added. Start with some value for heat addition
	 T0 = s% star_age  !Starting star age in years
 	 T1 = s% star_age  !Starting star age in years
 	 Mbh = 1d5*Msun !Black hole mass
	 
	 M = s% star_mass*Msun ! mass of WD in grams
         R = s% r(1) ! radius of WD in centimeters

	 eta=20
	 ecc=0.95
 
	 sma = R*eta/(1.-ecc)*(Mbh/M)**(1./3.) !semimajor axis
	 b0 = 64./5.*standard_cgrav**3*M*Mbh*(Mbh+M)/clight**5; !in cm^4/s
	 c0 = sma*(1.-ecc**2)/(1 + 121./304.*ecc**2)**(870./2299.)/(ecc**(12./19.))!in cm

	 prefac = -19./12.*b0/c0**4*secyer ! in yr^-1 (needed for orbital evolution)


         if (ierr /= 0) return
         extras_startup = 0
         call system_clock(time0,clock_rate)
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup
      
      
      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         double precision :: dt
         character (len=strlen) :: test
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         call system_clock(time1,clock_rate)
         dt = dble(time1 - time0) / clock_rate / 60
         call GET_ENVIRONMENT_VARIABLE( &
            "MESA_TEST_SUITE_CHECK_RUNTIME", test, status=ierr, trim_name=.true.)
         if (ierr == 0 .and. trim(test) == 'true' .and. dt > 1.5*expected_runtime) then
            write(*,'(/,a70,2f12.1,99i10/)') &
               'failed: EXCESSIVE runtime, prev time, retries, backups, steps', &
               dt, expected_runtime, s% num_retries, s% num_backups, s% model_number
         else
            write(*,'(/,a50,2f12.1,99i10/)') 'runtime, prev time, retries, backups, steps', &
               dt, expected_runtime, s% num_retries, s% num_backups, s% model_number
         end if
         ierr = 0
      end subroutine extras_after_evolve
      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         
         real(dp), parameter :: Blocker_scaling_factor_after_TP = 5d0
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         include 'formats'
         
         extras_check_model = keep_going
         
         if (abs(s% Blocker_scaling_factor - Blocker_scaling_factor_after_TP) < 1d-8) return
         
         if (s% center_he4 < 1d-4 .and. &
               any(s% burn_he_conv_region(1:s% num_conv_boundaries)) .and. &
               s% he_core_mass - s% c_core_mass <= s% TP_he_shell_max) then
            !write(*,1) 'set Blocker_scaling_factor = Blocker_scaling_factor_after_TP', Blocker_scaling_factor_after_TP
            !s% Blocker_scaling_factor = Blocker_scaling_factor_after_TP
            extras_check_model = terminate
         end if
         
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 5
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

	names(1) = 'ecc'
	vals(1) = ecc

	names(2) = 'eta'
	vals(2) = eta

	names(4) = 'Edot'
	vals(4) = Edot

	names(5) = 'Heat'
	vals(5) = Heat

	ierr = 0

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)
      
	M = s% star_mass*Msun ! mass of WD in grams
        R = s% r(1) ! radius of WD in centimeters	
	
	! If the time difference is more than 10 yrs, integrate. Otherwise, keep e,a
	Time =  s% star_age !current age in years
	tt = Time - T1 !Time difference between this step and the previous in years
	write(*,*) 'time elapsed', T1-T0
	write (*,*) 'timestep', tt
	estart=ecc;
	x1 = T1-T0;
	call rk4(estart, x1, tt/2.0_dp, estart) 
	x1 = T1-T0+tt*0.5_dp
	call rk4(estart, x1, tt/2.0_dp, estart)
	ecc = estart
	if (ecc<10d-4) extras_finish_step = terminate
        sma = c0*ecc**(12./19.)*(1 + 121.*ecc**2/304.)**(870./2299.)/(1.-ecc**2)
	eta = sma*(1.-ecc)*(M/Mbh)**(1./3.)/R
	T1 = Time;
	write(*,*) 'eccentricity', ecc
 	write(*,*) 'eta', eta

	!Hot WD	
	 Edota = -0.12610137375651373*ecc*ecc + 0.30701310595294512*ecc - 0.52874078635717681
	 Edotb = -2.3829938417811336*ecc*ecc + 4.5161596399810344*ecc - 9.8627895783988606

	!Cool WD
	!Edota = -0.11264803410705806*ecc**2 + 0.24202147481446407*ecc - 0.4004812884888681
	!Edotb = -2.1982156955629182*ecc**2 + 5.2265367124585449*ecc - 11.493252202747513
	

	if (eta < 10.) then	
		Edot = exp(Edota*10. + Edotb)
!		Jdot = 0.
	else
		Edot = exp(Edota*eta + Edotb)
!		Jdot = exp(Jdota*eta + Jdotb)
	endif
	write(*,*) 'Edot', Edot	

	Heat = (1.-ecc)**1.5*Edot/eta**7.5
	Heat = Heat*standard_cgrav**(1.5)*M**(2.5)/R**(2.5)
	!write(*,*) 'Heat', Heat
	Heat = Heat/M/(1d-4)
	write(*,*) 'Heat', Heat
	s% inject_uniform_extra_heat=Heat
	s% min_q_for_uniform_extra_heat = (1.-1d-4)

      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      	 if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
	end function extras_finish_step

      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            double precision :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
      
